"""
Classes and methods to load data from several sources into Deirokay.
"""

from typing import List, Type

from deirokay._typing import DeirokayReadCallable
from deirokay.backend import MultiBackendMixin
from deirokay.enums import Backend

from .dask_reader import read as read_dask
from .pandas_reader import read as read_pandas


class DataReader(MultiBackendMixin):
    """Helper class to load data from a data source in function of the
    backend."""

    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]
    read: DeirokayReadCallable


DataReader.register_backend_method(
    "read",
    staticmethod(read_dask),  # type: ignore  # Staticmethod is Callable
    Backend.DASK,
)
DataReader.register_backend_method(
    "read",
    staticmethod(read_pandas),  # type: ignore  # Staticmethod is Callable
    Backend.PANDAS,
)


def reader_factory(backend: Backend) -> Type[DataReader]:
    """Return the proper reader for the given backend.

    Parameters
    ----------
    backend : Backend
        Backend.

    Returns
    -------
    Type[DataReader]
        Subclass of DataReader containing a `read` staticmethod.
    """
    return DataReader.attach_backend(backend)


__all__ = ("reader_factory",)



"""
Set of functions related to Deirokay validation.
"""

import json
import warnings
from collections import deque
from copy import deepcopy
from datetime import datetime
from os.path import splitext
from typing import Optional, Union

from jinja2 import BaseLoader
from jinja2 import StrictUndefined as strict
from jinja2.nativetypes import NativeEnvironment

from deirokay._utils import check_columns_in_df_columns, render_dict
from deirokay.backend import detect_backend
from deirokay.enums import SeverityLevel
from deirokay.exceptions import ValidationError
from deirokay.fs import FileSystem, LocalFileSystem, fs_factory
from deirokay.history_template import get_series
from deirokay.statements.loader import statement_factory

from ._typing import (
    DeirokayDataSource,
    DeirokayValidationDocument,
    DeirokayValidationResultDocument,
)


def validate(
    df: DeirokayDataSource,
    *,
    against: Union[str, DeirokayValidationDocument],
    save_to: Optional[str] = None,
    save_format: str = None,
    current_date: Optional[datetime] = None,
    raise_exception: bool = True,
    exception_level: SeverityLevel = SeverityLevel.CRITICAL,
    template: Optional[dict] = None,
) -> DeirokayValidationResultDocument:
    """Validate a Deirokay DataFrame against a well-defined Validation
    Document.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame preferencially parsed by Deirokay `data_reader`.
    against : Union[str, dict]
        A `dict`-like Validation Document or a local/S3 path to a
        validation file in either YAML or JSON format.
    save_to : str, optional
        Path to folder where the validation result document will be
        saved to. An subfolder named the same as the validation
        document named will be created in this path, and the results
        will be saved inside following the
        `<current_date>.<save_format>` pattern. `<current_date>` is
        formatted as `%Y%m%dT%H%M%S` and `<save_format>` can be either
        `yaml` or `json`.
        If None, no validation log will be saved. By default None.
    save_format : str, optional
        Format in which to save the validation document
        (`yaml` or `json`).
        Only valid when `save_to` is not None.
        If None and `against` is a Python dictionary, defaults to YAML.
        If None and `against` is a valid path to a YAML or JSON file,
        it will keep this format.
        By default None
    current_date : datetime, optional
        Python `datetime.datetime` to use in log file name when saving
        validation result.
        Only valid when `save_to` is not None.
        If None, defaults to `datetime.utcnow()` value and raises a
        warning.
        By default None
    raise_exception : bool, optional
        Whether or not to raise a ValidationError exception whenever a
        statement whose level is greater or equal to `exception_level`
        fails.
        By default True
    exception_level : SeverityLevel, optional
        Minimum statement severity to raise exception for when
        statement validation fails.
        Only valid when `raise_exception` is True.
        By default SeverityLevel.CRITICAL (5).
    template : dict, optional
        Map of custom templates to be replaced in validation document
        before evaluation of statements. For mapped values, if
        callable, the returned value is used instead.

    Returns
    -------
    DeirokayValidationResultDocument
        Validation Result Document dict.

    Raises
    ------
    ValueError
        `save_to` parameter is not a directory neither an S3 path.
    ValidationError
        Validation failed for at least one statement whose severity is
        greater or equal to `exception_level`.
    """
    backend = detect_backend(df)

    if save_to:
        save_to_fs = fs_factory(save_to)
        if isinstance(save_to_fs, LocalFileSystem) and not save_to_fs.isdir():
            raise ValueError(
                "The `save_to` parameter must be an existing"
                " directory or an S3 path."
            )

    if isinstance(against, str):
        save_format = save_format or splitext(against)[1].lstrip(".")
        validation_document = fs_factory(against).read_dict()
    else:
        save_format = save_format or "yaml"
        validation_document = deepcopy(against)
    assert save_format.lower() in (
        "json",
        "yaml",
        "yml",
    ), f"Not a valid format {save_format}"

    # Render templates
    template = dict(
        series=lambda x, y: get_series(x, y, read_from=save_to_fs), **(template or {})
    )
    render_dict(
        NativeEnvironment(loader=BaseLoader(), undefined=strict),
        dict_=validation_document,
        template=template,
    )

    for item in validation_document["items"]:
        scope = item["scope"]
        scope = [scope] if not isinstance(scope, list) else scope
        check_columns_in_df_columns(scope, df.columns)

        df_scope = df[scope]

        for stmt in item.get("statements"):
            report = statement_factory(stmt, backend)(df_scope)
            if report["result"] is True:
                report["result"] = "pass"
            else:
                report["result"] = "fail"
            stmt["report"] = report

    if save_to:
        _save_validation_document(
            validation_document, save_to_fs, save_format, current_date
        )

    if raise_exception:
        raise_validation(validation_document, exception_level)

    return validation_document


def raise_validation(
    validation_result_document: dict, exception_level: SeverityLevel
) -> None:
    """Check for a validation result `dict` and raise a
    `ValidationError` exception whenever a statement whose severity
    level is greater or equal to `exception_level` fails.

    Parameters
    ----------
    validation_result_document : dict
        Validation Result Document generated by a Deirokay validation.
    exception_level : SeverityLevel
        Integer for the minimum severity level to raise exception for.

    Raises
    ------
    ValidationError
        Validation failed for at least one statement whose severity is
        greater or equal to `exception_level`.
    """
    highest_level = None
    for item in validation_result_document["items"]:
        scope = item["scope"]
        for stmt in item["statements"]:
            severity = stmt.get("severity", SeverityLevel.CRITICAL)
            result = stmt["report"]["result"]

            if result == "fail":
                if severity >= exception_level:
                    if highest_level is None or severity > highest_level:
                        highest_level = severity
                print(f"Statement failed for scope {scope}:")
                _sane_json_print(stmt)

    if highest_level is not None:
        print(f"Severity level threshold was {exception_level}.")
        raise ValidationError(
            highest_level, f"Validation failed with severity level {highest_level}."
        )


_TRUNCATE_OUTPUT_LIMIT = 64


def _sane_json_print(data: dict) -> None:
    """Print a dictionary to the stdout truncating the size of inner lists.

    Parameters
    ----------
    data : dict
        Dictionary object to be printed.
    """
    data = deepcopy(data)

    was_truncated = False
    nodes_to_visit = deque([(data, None, None)])

    def _process_iteratively():
        nonlocal was_truncated

        while nodes_to_visit:
            node, parent, index_key = nodes_to_visit.popleft()

            if isinstance(node, dict):
                for key, value in node.items():
                    if isinstance(value, (list, dict)):
                        nodes_to_visit.append((value, node, key))
            elif isinstance(node, list):
                if len(node) > _TRUNCATE_OUTPUT_LIMIT:
                    parent[index_key] = node[: _TRUNCATE_OUTPUT_LIMIT - 1] + ["..."]
                    was_truncated = True
                for index, value in enumerate(node):
                    if isinstance(value, (list, dict)):
                        nodes_to_visit.append((value, node, index))

    _process_iteratively()
    print(json.dumps(data, indent=4))

    if was_truncated:
        print("OBS: The output above was truncated to limit its size.")


def _save_validation_document(
    document: dict,
    save_to: FileSystem,
    save_format: Optional[str] = None,
    current_date: Optional[datetime] = None,
) -> None:
    if current_date is None:
        warnings.warn(
            "Document is being saved using the current date returned by the"
            " `datetime.utcnow()` method. Instead, prefer to explicitly pass a"
            " `current_date` argument to `validate`.",
            Warning,
        )
        current_date = datetime.utcnow()
    current_date = current_date.strftime("%Y%m%dT%H%M%S")  # type: ignore

    document_name = document["name"]
    folder_path = save_to / document_name
    if isinstance(folder_path, LocalFileSystem):
        folder_path.mkdir(parents=True, exist_ok=True)

    file_path = folder_path / f"{current_date}.{save_format}"

    print(f'Saving validation document to "{file_path!s}".')
    file_path.write_dict(document, indent=2)



from abc import ABC
from typing import Any, Iterable, List, Optional

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayDataSeries, DeirokaySerializedSeries
from deirokay.backend import MultiBackendMixin
from deirokay.enums import Backend, DTypes

from ..multibackend import treat as treat_


class BaseTreater(MultiBackendMixin, ABC):
    """Base class for all data treaters."""

    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]
    supported_dtype: Optional[DTypes] = None
    """Optional[DTypes]: DType treated by this treater."""
    supported_primitives: List[Any] = []
    """List[Any]: List of primitives supported by this treater."""

    def __call__(self, series: DeirokayDataSeries) -> DeirokayDataSeries:
        """Proxy for `treat`."""
        return self.treat(series)

    def treat(self, series: Iterable) -> DeirokayDataSeries:
        """Treat a raw Series to match data expectations for parsing
        and formatting.

        Calls the appropriate method for the detected backend

        Parameters
        ----------
        series : Iterable
            Raw series to be treated.
        """
        raise NotImplementedError

    @treat_(Backend.PANDAS, force=True)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        if isinstance(series, pandas.Series):
            return series
        return pandas.Series(series)

    @treat_(Backend.DASK, force=True)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        if isinstance(series, dask.dataframe.Series):
            return series

        # Assumption: When treating a general Iterable, we assume
        # it fits entirely in memory
        # TODO: let user propose a max partition size,
        # instead of assuming a single partition
        return dask.dataframe.from_pandas(pandas.Series(series), npartitions=1)

    @staticmethod
    def serialize(series: DeirokayDataSeries) -> DeirokaySerializedSeries:
        """Create a Deirokay-compatible serializable object that can
        be serialized (in JSON or YAML formats) and parsed back by
        Deirokay treaters.
        This method is useful when generating validation documents
        that embed Deirokay-compatible user data.
        The output format is a Python dict containing two keys:
        - `values`: list of values as Python object.
        - `parser`: Deirokay options to parse the data.
        (See more: `deirokay.parser.get_treater_instance`).

        Parameters
        ----------
        series : DeirokayDataSeries
            Series data to be serialized.

        Returns
        -------
        dict
            A Python dict containing the keys `values` and `parser`.
        """
        raise NotImplementedError



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

from typing import Iterable, Optional

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayDataSeries, DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .numeric_treater import NumericTreater


class FloatTreater(NumericTreater):
    """Treater for float variables

    Parameters
    ----------
    decimal_sep : Optional[str], optional
        Character to use as decimal separator, by default None
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.FLOAT64
    supported_primitives = [float]

    def __init__(self, decimal_sep: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        self.decimal_sep = decimal_sep

    def _treat_decimal_sep(self, series: DeirokayDataSeries) -> DeirokayDataSeries:
        if self.decimal_sep is not None and self.decimal_sep != ".":
            try:
                series = series.str.replace(self.decimal_sep, ".", regex=False)
            except AttributeError as e:
                print(*e.args)
                raise AttributeError(
                    "Make sure you are not declaring a `decimal_sep` to"
                    " read a non-text-like column. This may happen when"
                    " reading numeric columns from a .parquet file,"
                    " for instance."
                )
        return series

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        series = super()._treat_pandas(series)
        series = self._treat_decimal_sep(series)

        return series.astype(float).astype("Float64")

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        series = super()._treat_dask(series)
        series = self._treat_decimal_sep(series)

        return series.astype(float).astype("Float64")

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is pandas.NA:
                return None
            return float(item)

        return {
            "values": [_convert(item) for item in series],
            "parser": {"dtype": FloatTreater.supported_dtype.value},
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(series: "pandas.Series") -> DeirokaySerializedSeries:
        return FloatTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(series: "dask.dataframe.Series") -> DeirokaySerializedSeries:
        return FloatTreater._serialize_common(series)



"""
Functions to parse files into pandas DataFrames.
"""
import warnings
from copy import deepcopy
from typing import Optional, Union

from deirokay.__version__ import __comp_version__
from deirokay._typing import (
    DeirokayColumnOptions,
    DeirokayDataSource,
    DeirokayOptionsDocument,
)
from deirokay._utils import check_columns_in_df_columns
from deirokay.backend import detect_backend
from deirokay.enums import Backend
from deirokay.exceptions import InvalidBackend, ParsingError
from deirokay.fs import fs_factory

from .reader import reader_factory
from .treaters import get_treater_instance


def data_reader(
    data: Union[str, DeirokayDataSource],
    options: Union[str, DeirokayOptionsDocument],
    backend: Optional[Backend] = None,
    **kwargs,
) -> DeirokayDataSource:
    """Create a new tabular data from a file or an object and apply
    Deirokay treatments to correctly parse it and pre-validate its
    content.

    Parameters
    ----------
    data : Union[str, DeirokayDataSource]
        Path or object used as data source.
    options : Union[str, DeirokayOptionsDocument]
        Path to or `dict` representing a Deirokay Options document.
    backend: Optional[BackendValue], optional
        Defines backend to use for tables. By default None. Inferred
        from `data` when it is a valid DeirokayDataSource. Should be
        set when `data` is a path.

    Returns
    -------
    DeirokayDataSource
        A tabular data treated by Deirokay.
    """
    if isinstance(options, str):
        options_dict = fs_factory(options).read_dict()
    else:
        options_dict = deepcopy(options)
    options_dict.update(kwargs)
    columns = options_dict.pop("columns")

    if isinstance(data, str) and backend is None:
        if __comp_version__ < (2,):
            warnings.warn(
                "To preserve backward compatibility, the `backend` attribute"
                " is assumed to be `Backend.PANDAS` when reading data from a"
                " file or SQL query.\n"
                "In future, this behavior will change and an exception will"
                " be raised whenever the backend cannot be inferred from the"
                " `data` attribute. To prevent this error in future and"
                " suppress this warning in the current version,"
                " set the `backend` attribute explicitely in `data_reader()`.",
                FutureWarning,
            )
            backend = Backend.PANDAS
        elif __comp_version__ >= (2,):
            raise InvalidBackend(
                "You should provide a `backend` attribute when it cannot be"
                " inferred from `data`."
            )
    backend = backend or detect_backend(data)

    reader = reader_factory(backend)

    df = reader.read(data, columns=list(columns), **options_dict)
    data_treater(df, columns, backend)

    return df


def data_treater(
    df: DeirokayDataSource, options: DeirokayColumnOptions, backend: Backend
) -> None:
    """Receive options dict and call the proper treater class for each
    Deirokay data type.

    Parameters
    ----------
    df : DeirokayDataSource
        Raw tabular data to be treated.
    options : DeirokayColumnOptions
        Deirokay options.

    Raises
    ------
    NotImplementedError
        Data type not valid or not implemented.
    """
    check_columns_in_df_columns(options.keys(), df.columns)

    for col, opt in options.items():
        option: dict = opt.copy()

        try:
            df[col] = get_treater_instance(option, backend)(df[col])
        except Exception as e:
            raise ParsingError(f'Error when parsing "{col}".') from e



"""
Classes and methods to treat data immediately after being loaded
into Deirokay.
"""

from typing import Any, Type

from deirokay._typing import DeirokayOption
from deirokay._utils import recursive_subclass_generator
from deirokay.enums import Backend, DTypes

from .builtin import BaseTreater

DTYPE_2_TREATER = {
    cls.supported_dtype: cls
    for cls in recursive_subclass_generator(BaseTreater)
    if cls.supported_dtype is not None
}

PRIMITIVE_2_TREATER = {
    primitive: cls
    for cls in recursive_subclass_generator(BaseTreater)
    for primitive in cls.supported_primitives
}


def get_dtype_treater(dtype: Any) -> Type[BaseTreater]:
    """Map a dtype to its Treater class."""
    try:
        if isinstance(dtype, DTypes):
            return DTYPE_2_TREATER[dtype]
        elif isinstance(dtype, str):
            return DTYPE_2_TREATER[DTypes(dtype)]
        else:
            return PRIMITIVE_2_TREATER[dtype]

    except KeyError as e:
        raise NotImplementedError(
            f"Handler for '{dtype}' hasn't been" " implemented yet"
        ) from e


def get_treater_instance(option: DeirokayOption, backend: Backend) -> BaseTreater:
    """Create a treater instance from a Deirokay-style option.

    Example
    -------

    .. code-block:: python

        option = {
            'dtype': 'integer',
            'thousand_sep': ','
        }
    """
    option = option.copy()
    dtype = option.pop("dtype")

    cls = get_dtype_treater(dtype)
    return cls.attach_backend(backend)(**option)



"""
Module for FileSystem abstractions and utilities for multi-purpose
paths.
"""

import contextlib
import importlib
import itertools
import json
import os
import re
import sys
from collections import deque
from os.path import splitext
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import ModuleType
from typing import IO, Generator, Iterable, Optional, Sequence

import yaml
from typing_extensions import Literal

boto3_import_error = None
try:
    import boto3
except ImportError as e:
    boto3 = None
    boto3_import_error = e


BUCKET_KEY_REGEX = re.compile(r"s3:\/\/([\w\-]+)\/([\w\-\/.]*)")


def split_s3_path(s3_path: str) -> tuple:
    """Split a full s3 path into `bucket` and `key` parts.

    Parameters
    ----------
    s3_path : str
        Full S3 path, such as `s3://my-bucket/my-prefix/` or
        `s3://my-bucket/my-prefix/my-file.txt`.

    Returns
    -------
    tuple
        `(bucket, key)` tuple.
    """
    bucket, key = BUCKET_KEY_REGEX.findall(s3_path)[0]
    return bucket, key


def _import_file_as_python_module(path_to_file: str) -> ModuleType:
    """Import a .py file as a Python module.

    Parameters
    ----------
    path_to_file : str
        Path to .py file.

    Returns
    -------
    ModuleType
        Imported module.

    Raises
    ------
    ValueError
        Not a valid Python file.
    """

    module_path, extension = os.path.splitext(path_to_file)
    module_name = os.path.basename(module_path)

    if extension not in (".py", ".o"):
        raise ValueError("You should pass a valid Python file")

    module_dir = os.path.dirname(path_to_file)
    sys.path.insert(0, module_dir)
    module = importlib.import_module(module_name)
    sys.path.pop(0)

    return module


class FileSystem:
    """Abstract file system operations over folders and files in
    any place, such as local or in S3 buckets.

    Parameters
    ----------
    path : str
        Path to file or folder.
    """

    def __init__(self, path: str):
        self.path = path

    def ls(
        self,
        recursive: bool = False,
        files_only: bool = False,
        reverse: bool = False,
        limit: Optional[int] = None,
    ) -> Sequence["FileSystem"]:
        """List files in a prefix or folder.

        Parameters
        ----------
        recursive : bool, optional
            Whether or not to list subfolders recursively,
            by default False
        files_only : bool, optional
            List only files, ignore folders/prefixes,
            by default False.
        """
        raise NotImplementedError

    def read_dict(self, *args, **kwargs) -> dict:
        """Read and parse a `dict`-like file from either YAML or JSON
        format.

        Returns
        -------
        dict
            Python dictionary with the file content.
        """
        extension = splitext(self.path)[1].lower()
        if extension == ".json":
            return self.read_json(*args, **kwargs)
        elif extension in (".yaml", ".yml"):
            return self.read_yaml(*args, **kwargs)
        raise NotImplementedError(f"No parser for file type: {extension}")

    def write_dict(self, *args, **kwargs) -> None:
        """Serialize and write a Python `dict` to either YAML or JSON
        file."""
        extension = splitext(self.path)[1].lower()
        if extension == ".json":
            return self.write_json(*args, **kwargs)
        elif extension in (".yaml", ".yml"):
            return self.write_yaml(*args, **kwargs)
        raise NotImplementedError(f"No serializer for file type: {extension}")

    def read_yaml(self) -> dict:
        """Read and parse a YAML file as a Python `dict`.

        Returns
        -------
        dict
            Python dictionary with the file content.
        """
        with self.open("r") as fp:
            return yaml.safe_load(fp)

    def write_yaml(self, doc: dict, **kwargs) -> None:
        """Serialize and write a Python `dict` to a YAML file.

        Parameters
        ----------
        doc : dict
            The Python `dict`.

        """
        with self.open("w") as fp:
            yaml.dump(doc, fp, sort_keys=False, **kwargs)

    def read_json(self) -> dict:
        """Read and parse a JSON file as a Python `dict`.

        Returns
        -------
        dict
            Python dictionary with the file content.
        """
        with self.open("r") as fp:
            return json.load(fp)

    def write_json(self, doc: dict, **kwargs) -> None:
        """Serialize and write a Python `dict` to a JSON file.

        Parameters
        ----------
        doc : dict
            The Python `dict`.
        """
        with self.open("w") as fp:
            json.dump(doc, fp, **kwargs)

    def isdir(self) -> bool:
        """Return True if the path is a directory.

        Returns
        -------
        bool
            Whether or not the path is a directory

        Raises
        ------
        NotImplementedError
            Operation not valid or not implemented.
        """
        raise NotImplementedError

    def mkdir(self, *args, **kwargs):
        """Create directory using `Path.mkdir` method. Arguments are
        passed directly to this method.

        Raises
        ------
        NotImplementedError
            Operator not valid or not implemented.
        """
        raise NotImplementedError

    def import_as_python_module(self) -> ModuleType:
        """Import file as a Python module."""
        raise NotImplementedError

    def open(self, mode: Literal["r", "w"], *args, **kwargs) -> IO:
        """Open file."""
        raise NotImplementedError

    def read(self, *args, **kwargs) -> str:
        """Read a file as text."""
        with self.open(*args, mode="r", **kwargs) as fp:
            return fp.read()

    def __truediv__(self, rest: str) -> "FileSystem":
        """Create another FileSystem object by '/'-joining a FileSystem
        object with a string.

        Parameters
        ----------
        rest : str
            The rest of the file path.

        Returns
        -------
        FileSystem
            The same FileSystem subclass as the original object.

        Raises
        ------
        TypeError
            `rest` should be a `str`.
        """
        if isinstance(rest, str):
            cls = type(self)
            return cls(os.path.join(self.path, rest))
        raise TypeError()

    def __lt__(self, other: "FileSystem") -> bool:
        assert isinstance(self, type(other))
        return self.path.__lt__(other.path)

    def __str__(self) -> str:
        return self.path


class LocalFileSystem(FileSystem):
    """FileSystem wrapper for local files and folders."""

    # docstr-coverage:inherited
    def ls(
        self,
        recursive: bool = False,
        files_only: bool = False,
        reverse: bool = False,
        limit: Optional[int] = None,
    ) -> Sequence["LocalFileSystem"]:
        if recursive is False:
            raise NotImplementedError
        if files_only is False:
            raise NotImplementedError

        def _recursive_list():
            for parent, _, files in sorted(
                os.walk(self.path), key=lambda parent, *_: parent, reverse=reverse
            ):
                for file in sorted(files, reverse=reverse):
                    yield os.path.join(parent, file)

        limited_files = itertools.islice(_recursive_list(), limit)

        return [LocalFileSystem(path) for path in limited_files]

    # docstr-coverage:inherited
    def import_as_python_module(self) -> ModuleType:
        module = _import_file_as_python_module(self.path)
        return module

    # docstr-coverage:inherited
    def isdir(self) -> bool:
        return os.path.isdir(self.path)

    # docstr-coverage:inherited
    def mkdir(self, *args, **kwargs) -> None:
        return Path(self.path).mkdir(*args, **kwargs)

    # docstr-coverage:inherited
    def open(self, mode: Literal["r", "w"], *args, **kwargs) -> IO:
        return open(self.path, mode, *args, **kwargs)


class S3FileSystem(FileSystem):
    """FileSystem wrapper for objects stored in AWS S3."""

    LIST_OBJECTS_MAX_KEYS = 1000

    def __init__(
        self,
        path: Optional[str] = None,
        bucket: Optional[str] = None,
        prefix_or_key: Optional[str] = None,
        client: Optional["boto3.client"] = None,
    ):
        if boto3 is None:
            raise ImportError(
                "S3-backend requires `boto3` module to be"
                f" installed ({boto3_import_error})"
            )

        if bool(path) == bool(bucket) or bool(path) == bool(prefix_or_key):
            raise ValueError(
                "Either `path` or `(bucket, prefix_or_key)` should be"
                " passed (but not both)."
            )

        if bucket and prefix_or_key:
            final_bucket, final_prefix_or_key = bucket, prefix_or_key
            final_path = os.path.join(bucket, prefix_or_key)
        elif path:
            final_bucket, final_prefix_or_key = split_s3_path(path)
            final_path = path

        super().__init__(final_path)
        self.client = client or boto3.client("s3")
        self.bucket = final_bucket
        self.prefix_or_key = final_prefix_or_key

    # docstr-coverage:inherited
    def ls(
        self,
        recursive: bool = False,
        files_only: bool = False,
        reverse: bool = False,
        limit: Optional[int] = None,
    ) -> Sequence["S3FileSystem"]:
        if recursive is False:
            raise NotImplementedError
        if files_only is False:
            raise NotImplementedError

        # To handle requests containing over 1000 items, we need to
        # paginate through the results.
        max_keys = self.LIST_OBJECTS_MAX_KEYS
        page_iterable = self.client.get_paginator("list_objects_v2").paginate(
            Bucket=self.bucket,
            Prefix=self.prefix_or_key,
            PaginationConfig={"PageSize": max_keys},
        )
        chained_items: Iterable
        if not reverse:
            chained_items = (
                item["Key"]
                for page in page_iterable
                for item in page.get("Contents", [])
            )
        else:
            # Use deque with maxlen to exhaust the paginator and keep
            # only the last pages (those necessary to satisfy the
            # limit).
            pages_to_keep = (
                None if limit is None else (limit + max_keys - 2) // max_keys + 1
            )
            last_pages = deque(iter(page_iterable), maxlen=pages_to_keep)
            # Worst case scenario: the last page contains only one
            # item. Hence, from 2 to 1001, we need at least 2 pages,
            # and so on. Thus, the expression `(limit+998)//1000 + 1`
            chained_items = (
                item["Key"]
                for page in reversed(last_pages)
                for item in reversed(page.get("Contents", []))
            )

        chained_items = itertools.islice(chained_items, limit)
        return [
            S3FileSystem(bucket=self.bucket, prefix_or_key=key, client=self.client)
            for key in chained_items
        ]

    # docstr-coverage:inherited
    def import_as_python_module(self) -> ModuleType:
        _, extension = os.path.splitext(self.path)
        with NamedTemporaryFile(suffix=extension) as tmp_fp:
            self.client.download_file(self.bucket, self.prefix_or_key, tmp_fp.name)
            module = _import_file_as_python_module(tmp_fp.name)
            return module

    # docstr-coverage:inherited
    def open(self, mode: Literal["r", "w"], *args, **kwargs) -> IO:
        if mode == "r":
            body = self.client.get_object(Bucket=self.bucket, Key=self.prefix_or_key)[
                "Body"
            ]
            return contextlib.closing(body)  # type: ignore
        elif mode == "w":

            def _writable() -> Generator[IO, None, None]:
                with NamedTemporaryFile(mode, *args, **kwargs) as tmp_fp:
                    yield tmp_fp
                    tmp_fp.flush()
                    self.client.upload_file(
                        tmp_fp.name, self.bucket, self.prefix_or_key
                    )

            return contextlib.contextmanager(_writable)()  # type: ignore
        raise NotImplementedError(f"Mode '{mode}' not supported.")


def fs_factory(path: str) -> FileSystem:
    """Factory for FileSystem objects.

    Parameters
    ----------
    path : str
        Local or S3 path.

    Returns
    -------
    FileSystem
        Proper FileSystem object for the path provided.
    """
    if path.startswith("s3://"):
        return S3FileSystem(path)
    else:
        return LocalFileSystem(path)



"""
Deirokay enumeration classes.
"""

from enum import Enum, IntEnum, unique


@unique
class Backend(str, Enum):
    """
    The backend that the statement will be generated for.
    """

    PANDAS = "pandas"
    DASK = "dask"


class DTypes(str, Enum):
    """Deirokay data types."""

    INT64 = "integer"
    """Integer values, treated by
    :ref:`IntegerTreater<deirokay.parser.treaters.IntegerTreater>`."""
    INTEGER = INT64
    """Alias for INT64."""
    FLOAT64 = "float"
    """Float values, treated by
    :ref:`FloatTreater<deirokay.parser.treaters.FloatTreater>`."""
    FLOAT = FLOAT64
    """Alias for FLOAT64."""
    STRING = "string"
    """Text values, treated by
    :ref:`StringTreater<deirokay.parser.treaters.StringTreater>`."""
    STR = STRING
    """Alias for STRING."""
    DATETIME = "datetime"
    """Datetime values, treated by
    :ref:`DateTime64Treater<deirokay.parser.treaters.DateTime64Treater>`."""
    DT = DATETIME
    """Alias for DATETIME."""
    DATE = "date"
    """Date values, treated by
    :ref:`DateTreater<deirokay.parser.treaters.DateTreater>`."""
    TIME = "time"
    """Time values, treated by
    :ref:`TimeTreater<deirokay.parser.treaters.TimeTreater>`."""
    BOOLEAN = "boolean"
    """Boolean values, treated by
    :ref:`BooleanTreater<deirokay.parser.treaters.BooleanTreater>`."""
    BOOL = BOOLEAN
    """Alias for BOOLEAN."""
    DECIMAL = "decimal"
    """Decimal values, treated by
    :ref:`DecimalTreater<deirokay.parser.treaters.DecimalTreater>`."""


@unique
class SeverityLevel(IntEnum):
    """Deirokay named Severity levels."""

    MINIMAL = 1
    """Minimal severity validation error."""
    WARNING = 3
    """Validation warning."""
    CRITICAL = 5
    """Critical validation error."""



"""
Statement to check the number of rows in a scope.
"""
from typing import List

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayStatement
from deirokay.enums import Backend

from ..multibackend import profile, report
from .base_statement import BaseStatement


class RowCount(BaseStatement):
    """Check if the number of rows (or the number of of distinct rows)
    in a scope is between a minimum and maximum value.

    The available options are:

    * `min`: The minimum number of rows. If None, no minimum is
      enforced. Default: None.
    * `max`: The maximum number of rows. If None, no maximum is
      enforced. Default: None.
    * `distinct`: If True, check the number of distinct rows instead of
      the total number of rows. Default: False.

    Providing no `min` or `max` parameters, the statement will act only
    as a logger for its statistics.

    When counting the total number of rows (`distinct=False`), this
    statement may be applied to any scope of your DataFrame, since
    every column would have the same number of rows.
    By convention, you should apply it to a scope containing all the
    columns of your DataFrame.

    To count the number of (not-)null rows, you should use the `not_null`
    statement instead.
    To count the number of unique rows, use the `unique` statement.

    Examples
    --------
    * After some historial analysis of your data, you found that the
      number of rows is always greater or equal to than 42.
      You can declare the following validation item to represent this
      rule:

    .. code-block:: json

        {
            "scope": ["foo", "bar"],
            "statements": [
                {
                    "name": "row_count",
                    "min": 42
                }
            ]
        }

    * You have a table of daily transactions from all branches of a
      company. Not all branches have transactions for every day, and
      new branches may be added at any time. You want to ensure that
      the number of branches that appears in your data does not vary
      sharply downwards (below 5% of its 7-day historical average),
      which could be a sign of failure to receive transactions from
      some branches.
      You can declare the following validation item (in YAML format) to
      check this rule:

    .. code-block:: yaml

        scope: branch_name
        statements:
        - name: row_count
          distinct: True
          min: >
            {{ 0.95 * (
              series("transactions", 7).branch_name.row_count.distinct_rows.mean()
              | default(19, true))
              | float
            ) }}

    There are many things going on here:

    * In YAML, the ">" operator is used to collapse a multi-line string
      into a single line. In JSON you would have to put everything in
      the same line;
    * The "{{}}" braces are used to indicate that the following
      expression is a Jinja2 template.
    * The `series` function is a built-in Deirokay method
      used to get the 7-day historical validations.
      Further down, the `mean` function is used to
      compute the 7-day average of the `distinct_rows` metric returned
      by the `row_count` statement in the `branch_name` scope.
    * The "|" operator inside the Jinja2 template is used to apply
      a function to the result of the previous expression, such that in
      the end we obtain a float.
    * The `default` function is used to set a default value if the
      previous expression is `None`.
    * The `float` function is used to convert the result of the
      previous expression, which is a `numpy.float64`, to a float,
      which can be properly serialized in JSON or YAML format when the
      validation logs are generated.

    For this example to work, you will need to declare in your
    `deirokay.validate` call the `save_to` parameters, so that the
    validation logs can be saved and later used to provide historical
    analysis.

    .. code-block:: python

        from deirokay import validate

        validate(df, against=assertions, save_to='logs')

    """

    name = "row_count"
    expected_parameters = ["min", "max", "distinct"]
    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.min = self.options.get("min", None)
        self.max = self.options.get("max", None)
        self.distinct = self.options.get("distinct", False)

    def _report_common(self, df):
        row_count = len(df)
        distinct_count = len(df.drop_duplicates())

        return {
            "rows": row_count,
            "distinct_rows": distinct_count,
        }

    @report(Backend.PANDAS)
    def _report_pandas(self, df: "pandas.DataFrame") -> dict:
        return self._report_common(df)

    @report(Backend.DASK)
    def _report_dask(self, df: "dask.dataframe.DataFrame") -> dict:
        return self._report_common(df)

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        if self.distinct:
            count = report["distinct_rows"]
        else:
            count = report["rows"]

        if self.min is not None:
            if not count >= self.min:
                return False
        if self.max is not None:
            if not count <= self.max:
                return False
        return True

    @staticmethod
    def _profile_common(df):
        statement: DeirokayStatement

        if len(df.columns) > 1:
            count = len(df)
            statement = {"type": "row_count", "min": count, "max": count}
        else:
            count = len(df.drop_duplicates())
            statement = {
                "type": "row_count",
                "distinct": True,
                "min": count,
                "max": count,
            }
        return statement

    @profile(Backend.PANDAS)
    @staticmethod
    def _profile_pandas(df: "pandas.DataFrame") -> DeirokayStatement:
        return RowCount._profile_common(df)

    @profile(Backend.DASK)
    @staticmethod
    def _profile_dask(df: "dask.dataframe.DataFrame") -> DeirokayStatement:
        return RowCount._profile_common(df)



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

import datetime
from typing import Iterable

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .validator import Validator

_supported_primitives = [datetime.datetime]
try:
    _supported_primitives.append(pandas.Timestamp)
except ModuleNotFoundError:
    pass


class DateTime64Treater(Validator):
    """Treater for datetime variables

    Parameters
    ----------
    format : str, optional
        Format to parse dates from, by default '%Y-%m-%d %H:%M:%S'
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.DATETIME
    supported_primitives = _supported_primitives

    def __init__(self, format: str = "%Y-%m-%d %H:%M:%S", **kwargs):
        super().__init__(**kwargs)

        self.format = format

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        series = super()._treat_pandas(series)

        return pandas.to_datetime(series, format=self.format)

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        series = super()._treat_dask(series)

        return dask.dataframe.to_datetime(series, format=self.format)

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is None or item is pandas.NaT:
                return None
            return str(item)

        return {
            "values": [_convert(item) for item in series],
            "parser": {"dtype": DateTime64Treater.supported_dtype.value},
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(series: "pandas.Series") -> DeirokaySerializedSeries:
        return DateTime64Treater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(series: "dask.dataframe.Series") -> DeirokaySerializedSeries:
        return DateTime64Treater._serialize_common(series)



"""
Implementation of builtin Deirokay statements.
"""
from deirokay._utils import recursive_subclass_generator

from .base_statement import BaseStatement
from .column_expression import ColumnExpression  # noqa: F401
from .contain import Contain  # noqa: F401
from .not_null import NotNull  # noqa: F401
from .row_count import RowCount  # noqa: F401
from .statistic_in_interval import StatisticInInterval  # noqa F401
from .unique import Unique  # noqa: F401

__all__ = tuple(cls.__name__ for cls in recursive_subclass_generator(BaseStatement))



from deirokay._typing import GeneralDecorator
from deirokay.backend import register_backend_method
from deirokay.enums import Backend


def treat(backend: Backend, *, force: bool = False) -> GeneralDecorator:
    """Define a decorator that turns any method into an alias for the
    `treat` method when executed using a given `backend`.

    Parameters
    ----------
    backend : Backend
        The backend for the decorated method.

    Returns
    -------
    GeneralDecorator
        A decorator for the alias method of `treat`.
    force : bool, optional
        Force overwrite target method when it already exists.
        Defaults to False.
    """
    return register_backend_method("treat", backend, force=force)


def serialize(backend: Backend, *, force: bool = False) -> GeneralDecorator:
    """Define a decorator that turns any method into an alias for the
    `serialize` method when executed using a given `backend`.

    Parameters
    ----------
    backend : Backend
        The backend for the decorated method.

    Returns
    -------
    GeneralDecorator
        A decorator for the alias method of `serialize`.
    force : bool, optional
        Force overwrite target method when it already exists.
        Defaults to False.
    """
    return register_backend_method("serialize", backend, force=force)



from os.path import dirname, join

from setuptools import find_packages, setup

with open(join(dirname(__file__), 'deirokay', '__version__.py')) as v:
    __version__ = None
    exec(v.read().strip())

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = [line.strip() for line in f.readlines()]

with open('requirements-dev.txt') as f:
    requirements_dev = [line.strip() for line in f.readlines()]

setup(
    name="deirokay",
    packages=find_packages(include=['deirokay*']),
    version=__version__,
    author="Marcos Bressan",
    author_email="marcos.bressan@bigdata.com.br",
    description="A tool for data profiling and data validation",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/bigdatabr/deirokay",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    include_package_data=True,
    zip_safe=True,
    install_requires=requirements,
    extras_require={
        'dev': requirements_dev,
        's3': ['boto3', 's3fs'],
        'pandas': ['pandas<2,>=1.3'],
        'dask': ['dask[dataframe]>=2022.1.1'],
    }
)



"""
Classes and methods to load and parse correct data into Deirokay.
"""

from .loader import data_reader
from .treaters import get_dtype_treater, get_treater_instance

__all__ = (
    "data_reader",
    "get_dtype_treater",
    "get_treater_instance",
)



"""
Implementation of builtin Deirokay treaters.
"""
from deirokay._utils import recursive_subclass_generator

from .base_treater import BaseTreater
from .boolean_treater import BooleanTreater  # noqa: F401
from .date_treater import DateTreater  # noqa: F401
from .datetime64_treater import DateTime64Treater  # noqa: F401
from .decimal_treater import DecimalTreater  # noqa: F401
from .float_treater import FloatTreater  # noqa: F401
from .integer_treater import IntegerTreater  # noqa: F401
from .numeric_treater import NumericTreater  # noqa: F401
from .string_treater import StringTreater  # noqa: F401
from .time_treater import TimeTreater  # noqa: F401
from .validator import Validator  # noqa: F401

__all__ = tuple(cls.__name__ for cls in recursive_subclass_generator(BaseTreater))
